#
# Sensus Consulting Ltd. (C) 1997-1998
# Matt Newman <matt@sensus.org>
#
# Generic Tree
#
#INDEX\
proc tcl++::tree {} {}

package require tcl++

proc tree {{this #auto} args} {uplevel 1 tcl++::tree $this $args}

class tcl++::tree {
	constructor {args}
    #
    # Get & Set attributes of specific nodes.
    #
    public {
	method get {id args}
	method put {id fld val args}

	method children {id {order -increasing}}
	method parent {id}

	method add {id args}
	method delete {id}
	method lookup {path}
	method path {id {sep /}}
	method walk {idRef body}
	#
	# Sort by list of fields.
	# field may be pair, where first no is "asc" or "desc" for
	# ascending or decending sort order.
	method sort {list args}
	#
	# Convert subtree indicated by id into an XML data-stream.
	#
	method export {{id 0}}
	#
	# Populate this tree, below $id with the packed tree in $data
	#
	method import {data {id 0} {skip 1}}
	#
	# Debug & Convenience routines
	#
	method print {}
    }
    protected {
    	method sortPair {flds id1 id2} 

	variable uid 0
	variable store
    }
	#
	# Private methods
	#
	private method _lookup1 {id name}
}

body tcl++::tree::constructor {args} {
	set id 0

	set store($id) [list id $id name $id class index]
	set store(${id},parent)		$id
	set store(${id},children)	{}

	if {[llength $args]>0} {
		set store($id) [eval pset [list $store($id)] $args]
	}
}
#
# Get & Set attributes of specific nodes.
#
body tcl++::tree::get {id args} {
	switch -- [llength $args] {
	0	{return $store($id)}
	1	-
	2	{return [eval [list pget $store($id)] $args]}
	default	{
		error "wrong # of args: should be \"get id ?fld ?default??\""
	}
	};# switch
}
body tcl++::tree::put {id fld val args} {
	set store($id) [eval pset [list $store($id) $fld $val] $args]
}
body tcl++::tree::children {id {order -increasing}} {
	return $store(${id},children)
}
body tcl++::tree::parent {id} {
	return $store(${id},parent)
}
body tcl++::tree::add {id args} {
	set cid [incr uid]

	lappend store(${id},children) $cid

	set store($cid) [list id $cid name "" class node]
	set store(${cid},parent)	$id
	set store(${cid},children)	{}

	if {[llength $args]>0} {
		set store($cid) [eval pset [list $store($cid)] $args]
	}
	return $cid
}
body tcl++::tree::delete {id} {
	foreach cid $store(${id},children) {
		delete $cid
	}
	set pid $store(${id},parent)
	set idx [lsearch $store(${pid},children) $id]
	set store(${pid},children) [lreplace $store(${pid},children) $idx $idx]

	unset store($id)
}
body tcl++::tree::path {id {sep /}} {
	set name [get $id name]
	set pid [parent $id]

	if {$pid == $id} {
		return ""
	} else {
		set path [path $pid $sep]
		if {$path == ""} {
			return $name
		} else {
			return "$path$sep$name"
		}
	}
}
body tcl++::tree::lookup {path} {
	set id 0
	set p [split $path /]
	while {![lempty $p]} {
		set id [_lookup1 $id [lvarpop p]]
		if {$id == ""} {
			error "$path not found"
		}
	}
	return $id
}
body tcl++::tree::sort {list args} {
	if {$list == "node"} {
		set id [lindex $args 0]
		set args [lrange $args 1 end]
		set list $store($id,children)
		set store($id,children) [lsort -command [code $this sortPair $args] $list]
	} else {
		return [lsort -command [code $this sortPair $args] $list]
	}
}
body tcl++::tree::sortPair {flds id1 id2} {
	set m(asc) 1
	set m(desc) -1
	set dir asc
	foreach fld $flds {
		if {[llength $fld]>1} {
			set dir [lindex $fld 0]
			set fld [lindex $fld 1]
		}
		set v1 [pget $store($id1) $fld ""]
		set v2 [pget $store($id2) $fld ""]
		if {$v1 < $v2} {
			return [expr {$m($dir) * -1}]
		} elseif {$v1 > $v2} {
			return [expr {$m($dir) * 1}]
		}
	}
	return 0
}
body tcl++::tree::walk {idRef body} {
	upvar 1 $idRef id

	set stack [children 0 -decreasing]

	upvar 1 _error_ ret
	while {![lempty $stack]} {
		set id [lvarpop stack]
		set retCode [uplevel 1 [list catch $body _error_]]
		switch -- $retCode {
		0	{#TCL_OK}
		2	{#TCL_RETURN
			return -code return $ret
			}
		3	{#TCL_BREAK
			break
			}
		4	{#TCL_CONTINUE
			continue
			}
		1	-
		default	{#TCL_ERROR
			global errorInfo errorCode
			return -code $retCode \
				-errorinfo $errorInfo \
				-errorcode $errorCode $ret
			}
		}
		foreach cid [[virtual]::children $id -decreasing] {
			lvarpush stack $cid
		}
	}
}
#
# Convert subtree indicated by id into a data-stream.
#
body tcl++::tree::export {{id 0}} {
	set tag [pget $store($id) class]
	set attrs [pdel $store($id) id]
	set attrs [pdel $attrs class]
	set attrs [eval xml::putattrs $attrs]

	if {[lempty $store(${id},children)]} {
		set data "<${tag} $attrs />\n"
	} else {
		set data "<${tag} $attrs >\n"
		foreach cid $store(${id},children) {
			append data [export $cid]
		}
		append data "</${tag}>\n"
	}
	return $data
}
#
# Populate this tree, below $id with the packed tree in $data
#
body tcl++::tree::import {data {id 0} {skip 1}} {
	set data [xml::parse $data]

	set n 0
	set stack {}
	foreach {tag close empty param text} $data {
		if {$tag == ""} continue
		if {$skip} {
			set skip 0
			continue
		}
		if {[string length $close]!=0} {
			set id [lvarpop stack]
			continue
		}
		set cmd [list add $id class $tag]
		foreach {opt val} [xml::getattrs $param] {
			lappend cmd $opt $val
		}
		set cid [eval $cmd]
		incr n
		if {[string length $empty]==0} {
			lvarpush stack $id
			set id $cid
		}
	}
	if {[llength $stack]!=0} {
		error "unmatched tags: stack=$stack"
	}
	return $n
}
#
# Debug & Convenience routines
#
body tcl++::tree::print {} {
	walk id {
		puts stdout [path $id]
		#update idletasks
	}
}
#
# Private methods
#
body tcl++::tree::_lookup1 {id name} {
	foreach cid $store(${id},children) {
		set cname [get $cid name]
		if {[string compare $name $cname]==0} {
			return $cid
		}
	}
	return ""
}
